<?
  require "../relation_types/relation_type.php";

  /* helper classes */
  class Node {
    var $nodeId;
    var $nodeName;
    var $subNodes;
    
    /* iterator support */
    var $pos;
    
    function Node($nodeId, $nodeName) {
      $this->nodeId = $nodeId;
      $this->nodeName = $nodeName;
      $this->subNodes = Array();
    }
    
    function addSubNode($subNode) {
      $this->subNodes[] = $subNode;
    }
    
    function resetIter() {
      $this->pos = 0;
    }
    
    function nextSubNode() {
      if($this->pos >= Count($this->subNodes)) {
        return false;
      }
      else {
        return ($this->subNodes[$this->pos++]);
      }
    }
    
    function currentSubNode() {
      if($this->pos >= Count($this->subNodes)) {
        return false;
      }
      else {
        return ($this->subNodes[$this->pos]);
      }
    }
  }
  
  class Stack {
    var $elements;
   
    function Stack() {
     $this->elements = Array();
    }
   
    function push($element) {
      $this->elements[Count($this->elements)] = $element;
    }
   
    function pop() {
      $count = Count($this->elements);
      $element = $this->elements[$count-1];
      UnSet($this->elements[$count-1]); //remove top element
      return $element;
    }
   
    function size() {
      return Count($this->elements);
    }
    
/* "class Node" specific debugging feature
    function printStack() {
      $cnt = Count($this->elements);
      echo "<p>Stack print<br>\n";
      for($i=0; $i<$cnt; $i++) {
        $el = $this->elements[$i];
        echo "$el->nodeId (".$el->currentSubNode().")<br>\n";
      }
    }
*/    
  }

  /* dimension class */
  class Dimension {
    var $conn;  //connection handle
    
    var $oid;
    var $relid;
    var $name;

    /* internals */
    var $nodes;  //map of nodes
    var $stk;    //stack object
    var $wr;    //walk reset flag
        
    function Dimension($conn, $oid, $relid=0) {
      $this->conn = $conn;
      $this->oid = $oid;
      $this->relid = $relid;
      if($this->relid == 0) {//invalid value
        $this->relid = $this->getBaseRelationID();
      }
      
      $this->nodes = null;
      $this->stk = null;
      $this->wr = false;
    }
    
    function getRootID() {
      $command = "SELECT HATR_ID FROM H_ATRT_VALUES WHERE ATRT_ID=".$this->oid;
      $command .= " AND ATRT_VALUE='root'";
      
      //debug_trace($command);
      $rs = ODBC_Exec($this->conn, $command);
      if(!ODBC_Fetch_Row($rs))
        return false;
      else
        return ODBC_Result($rs, 1);
    }
    
    function getBaseRelationID() {
      $command = "SELECT RLN_ID FROM RELATION_TYPES WHERE ATRT_ID_L = ATRT_ID_R AND ATRT_ID_L =".$this->oid;
      $command .= " AND RLN_KIND = 'L' AND RLN_NAME = RLV_NAME AND RLN_NAME='base'";
      
      $rs = ODBC_Exec($this->conn, $command);
      ODBC_Fetch_Row($rs);
      return ODBC_Result($rs, 1);
    }
    
    function create() {
        $br = new RelationType($this->conn);
        $br->oid = $br->newOid();
         $br->name = "base";
        $br->revName = "base";
        $br->atrt_id_l = $this->oid;
        $br->atrt_id_r = $this->oid;
        $br->kind = "L";
        $br->r = "Y";
        $br->s = "N";
        $br->t = "Y";
        $br->create();
        
        /* create "root" element */
        $rs = ODBC_Exec($this->conn, "SELECT Count(HATR_ID) as _COUNT, Max(HATR_ID) as _MAX FROM H_ATRT_VALUES");
        ODBC_Fetch_Row($rs);
        $count = ODBC_Result($rs, 1);
        $max =  ODBC_Result($rs, 2);
        $hid = 0;
      
        if($count == 0)
          $hid = 1;
        else
          $hid = ($max + 1);
        
        $command = "INSERT INTO H_ATRT_VALUES (HATR_ID, ATRT_ID, ATRT_VALUE) VALUES (";
        $command .= $hid;
        $command .= ", ".$this->oid;
        $command .= ", 'root')";
        
        //debug_trace($command);
        ODBC_Exec($this->conn, $command);
    }
    
    function delete() {
      $br = new RelationType($this->conn);
      $br->oid = $this->relid;
      $br->delete();
      
      $command = "DELETE FROM H_ATRT_VALUES WHERE ATRT_ID=".$this->oid;
      
      ODBC_Exec($this->conn, $command);
    }
    
    function loadDimension() {
      /* load name */
      $command = "SELECT ATRT_NAME FROM ATTRIBUTE_TYPES WHERE ATRT_ID=".$this->oid;
      
      $rs = ODBC_Exec($this->conn, $command);
      ODBC_Fetch_Row($rs);
      $this->name = ODBC_Result($rs, 1);
    
      /* load nodes */    
      $this->nodes = Array(); //re-initialize
      
      $command = "SELECT HATR_ID, ATRT_VALUE FROM H_ATRT_VALUES";
      $command .= " WHERE ATRT_ID=".$this->oid;
      
      $rs = ODBC_Exec($this->conn, $command);
      while(ODBC_Fetch_Row($rs)) {
        $hid = (string)ODBC_Result($rs, 1);
        $name = ODBC_Result($rs, 2);
        
        $nd = new Node($hid, $name);
        $this->nodes[$hid] = $nd;
      }
    
      /* load node relations (structure) */
      $command = "SELECT HATR_ID_L, HATR_ID_R";
      $command .= " FROM RELATIONS, H_ATRT_VALUES";
      $command .= " WHERE RLN_ID=".$this->relid;
      $command .= " AND HATR_ID_R = HATR_ID";       //join with values table
      $command .= " ORDER BY HATR_ID_L, ATRT_VALUE";

      $rs = ODBC_Exec($this->conn, $command);
      while(ODBC_Fetch_Row($rs)) {
        $left = (string)ODBC_Result($rs, 1);
        $right = (string)ODBC_Result($rs, 2);
        $nd = $this->nodes[$left];  //read from array
        $nd->addSubNode($right);    //change
        $this->nodes[$left] = $nd;  //write back
      }
    }
    
    function resetWalk() {
      $this->stk = new Stack(); //initialize a new Stack
      $this->wr = true;          //walk reset => begin with "root"
    }
    
    /* helper function for getNextElement() */
    function pushNode($nid, &$level, &$nodeid, &$nodename) {
      //debug_trace("pushNode/\$nid=".$nid);
      $nd = $this->nodes[$nid];
      
      $nd->resetIter();    //reset to the first element
      $this->stk->push($nd);

      $level = $this->stk->size();
      $nodeid = $nd->nodeId;
      $nodename = $nd->nodeName;
    }
    
    /* helper function for getNextElement() */
    function nextNodeId() {
      if($this->stk->size() == 0) {
        return false;
      }
      $nd = $this->stk->pop();
      //$this->stk->printStack();
      $nodeid = $nd->nextSubNode();
      //debug_trace("|".$nodeid."|");
      if($nodeid == false) {  //last subNode reached
        //debug_trace("recur");
        return $this->nextNodeId(); //call itself recursively
      }
      else {
        //debug_trace("next_el");
        $this->stk->push($nd);
        return $nodeid;
      }
    }
    
    /* if succeeds returns true, false otherwise */
    function getNextElement(&$level, &$nodeid, &$nodename) {
      //$this->stk->printStack();
      if($this->wr == true) {  //first node, special care
        $this->wr = false;
        $rid = (string)($this->getRootID());
        $this->pushNode($rid, $level, $nodeid, $nodename);
        
        return true;
      }
      else {
        $nid = $this->nextNodeId();
//        debug_trace("\$nid=".$nid);
        if($nid == false) {
          return false;
        }
        else {
          $this->pushNode($nid, $level, $nodeid, $nodename);
          return true;
        }
      }
    }
  }
?>