<?
  
  /*
    contains:
    - Coinstraint class
    - CoinstraintExpr class
    - CoinstraintSeq  class
  */

  class Constraint {
    var $conn;
      
    var $not;      //"NOT" or ""
    var $atrt_id;  //attribute type id
    var $hie;       //true if $atrt_id is hierarchical
    var $rev;       //reverse relation flag (only if $hie == true)
    var $rln;       //ordop ("=", ...) or RLN_ID (depends on $hie)
    var $atrt_val; //value or HATR_ID (depends on $hie)

    function Constraint($conn) {
      $this->conn = $conn;
    }

    /* 
        $cont - input
        $rlns - set of used relation (for validation check)
    */
    function parse($constr, &$rlns, $otid) {
      $len = StrLen($constr);
      if($constr[$len-1] != "\"") {
        syn_err(sprintf("Value must be enclosed in quotation marks in [%s]", $constr));
        return false;
      }
      $cont = SubStr($constr, 0, $len-1); //cut ending quote
      $parts = explode("\"", $cont); //separate
      
      $words = split("[ ]+", Trim($parts[0]));
      $value = $parts[1];
//      debug_trace("constraint: ".$cont);
//      print_array($words);
//      debug_trace("value: ".$value);
      
      $not;
      $atname;
      $rev;
      $op;
      /* 2 - 4 words */
      switch(Count($words)) {
        case 2: //atribute name, operator
          $not = "";
          $atname = $words[0];
          $rev = false;
          $op = $words[1];
          break;
        case 3: //not, attribute name, operator | attribute name, rev, operator
          if(StrToUpper($words[0]) == "NOT") {
            $not = "NOT";
            $atname = $words[1];
            $rev = false;
            $op = $words[2];
          }
          else if(StrToUpper($words[1]) == "REV") {
            $not = "";
            $atname = $words[0];
            $rev = true;
            $op = $words[2];
          }
          else {
            syn_error(sprintf("Invalid constraint syntax [%s]", $constr));
            return false;
          }
          break;
        case 4: //not, attribute name, rev, operator
          $not = $words[0];
          $atname = $words[1];
          $rev = $words[2];
          $op = $words[3];
          if(StrToUpper($not) != "NOT") {
            syn_error(sprintf("Invalid constraint syntax [%s], NOT operator [%s]", $constr, $not));
            return false;
          }
          else
            $not = "NOT";
          if(StrToUpper($rev) != "REV") {
            syn_error(sprintf("Invalid constraint syntax [%s], REV operator [%s]", $constr, $rev));
            return false;
          }
          else
           $rev = true; //set to boolean value
          break;
        default:
          syn_error(sprintf("Invalid constraint syntax [%s]", $constr));
          return false;
      }
      
      // pre-process and check values
      $atrt_id = getAtrtId($this->conn, $atname, $hie);
      if(!$atrt_id)
        return false;
      if($hie &&  isOrdOp($op)) {
        sem_err(sprintf("Cannot use operator [%s] with domain attribute [%s]", $op, $atname));
        return false;
      }
      if($hie == false && !isOrdOp($op)) {
        sem_err(sprintf("Cannot use relation [%s] with ordinal attribute [%s]", $op, $atname));
        return false;
      }
      if(isOrdOp($op) && $rev) {
        sem_err(sprintf("Cannot use REV with operator [%s]", $op));
        return false;
      }
      if(!objHasAttribute($this->conn, $otid, $atrt_id)) {
        sem_err(sprintf("Object [%s] does not have attribute [%s]!", 
          getTypeName($this->conn, $otid), $atname));
        return false;
      }
      if($hie) {
        $rln_id = getRelationId($this->conn, $op, $rev, $atrt_id, $atname);
        if(!$rln_id)
          return false;
        $hatr_id = getHatrId($this->conn, $atrt_id, $atname, $value);
        if(!hatr_id)
          return false;
        $this->not = $not;
        $this->atrt_id = $atrt_id;
        $this->hie = $hie;
        $this->rev = $rev;
        $this->rln = $rln_id;
        $this->atrt_val = $hatr_id;

        // add relation to the "must validate" list
        $rlns[$this->rln] = 1;
      }
      else {
        $this->not = $not;
        $this->atrt_id = $atrt_id;
        $this->hie = $hie;
        $this->rev = $rev;
        $this->rln = $op;
        $this->atrt_val = $value;
      }
      
//      debug_trace("Constraint parsed: [".$constr."]");
//      debug_trace(sprintf("not: [%s] atrt_id: [%s] hie: [%s] rev: [%s] rln: [%s] atrt_val: [%s]",
//        $this->not, $this->atrt_id, $this->hie, $this->rev, $this->rln, $this->atrt_val));
      return true;
    }
    
    function dumpParsed() {
      return sprintf("%s%s{%s} %s%s \"%s\"",
        $this->not ? "NOT " : "", 
        $this->atrt_id, 
        $this->hie ? "H" : "O",
        $this->rev ? "REV " : "",
        $this->rln,
        $this->atrt_val);
    }    

    function buildSelect($uplink, &$cnt, &$from, &$where) {
      if($this->hie) { //hierarchical operator constraint
        $hovta = "HOV".($cnt++);
        $from .= ", H_OBJ_VALUES ".$hovta;
        $rlnta = "RLNC".($cnt++);
        $from .= ", RELATIONS_CACHE ".$rlnta;
        //subselect table (needed for "NOT" variant)
        $rlnt2 = "RLNC".($cnt++);
        
        $where .= sprintf("%s.OID=%s.OID AND %s.ATRT_ID=%d AND %s.HATR_ID=",
          $uplink, $hovta, $hovta, $this->atrt_id, $hovta);
        if($this->rev) {
          $where .= sprintf("%s.HATR_ID_L AND %s.RLN_ID=%d AND ",
            $rlnta, $rlnta, $this->rln);
          $where .= sprintf("%s.HATR_ID_L %s IN (SELECT %s.HATR_ID_L FROM RELATIONS_CACHE %s WHERE %s.HATR_ID_R=%d)",
            $rlnta, $this->not, $rlnt2, $rlnt2, $rlnt2, $this->atrt_val);
        }
        else {
          $where .= sprintf("%s.HATR_ID_R AND %s.RLN_ID=%d AND ",
            $rlnta, $rlnta, $this->rln);
          $where .= sprintf("%s.HATR_ID_R %s IN (SELECT %s.HATR_ID_R FROM RELATIONS_CACHE %s WHERE %s.HATR_ID_L=%d)",
            $rlnta, $this->not, $rlnt2, $rlnt2, $rlnt2, $this->atrt_val);
        }
      }
      else { //ordinal operator constraint
        $oovta = "OOV".($cnt++);
        $from .= ", O_OBJ_VALUES ".$oovta;
        $where .= sprintf("%s.OID=%s.OID AND %s.ATRT_ID=%d AND %s %s.ATRT_VALUE%s'%s'",
          $uplink, $oovta, $oovta, $this->atrt_id, $this->not, $oovta, $this->rln, $this->atrt_val);
      }
      return true;
    }
  }

  class ConstraintExpr {
    var $conn;
      
    var $cont;  //a single constraint
    var $cseq;  //a constraint sequence

    function ConstraintExpr($conn) {
      $this->conn = $conn;
    }

    /* 
        $cexp - input
        $rlns - set of used relation (for validation check)
    */
    function parse($cexp, &$rlns, $otid) {
      if(eregi("[ \)]AND[ \(]", $cexp) || eregi("[ \)]OR[ \(]", $cexp)) {
        $cseq = new ConstraintSeq($this->conn);
        if(!$cseq->parse($cexp, $rlns, $otid))
          return false;
        $this->cseq = $cseq;
        $this->cont = false;
      }
      else {
        $cont = new Constraint($this->conn);
        if(!$cont->parse($cexp, $rlns, $otid))
          return false;
        $this->cseq = false;
        $this->cont = $cont;
      }
      return true;
    }
    
    function dumpParsed() {
      if($this->cseq)
        return "(".$this->cseq->dumpParsed().")";
      else
        return "(".$this->cont->dumpParsed().")";
    }

    function buildSelect($uplink, &$cnt, &$from, &$where) {
      if($this->cseq) {
        $where .= "(";
        $this->cseq->buildSelect($uplink, $cnt, $from, $where);
        $where .= ")";
      }
      else {
        $where .= "(";
        $this->cont->buildSelect($uplink, $cnt, $from, $where);
        $where .= ")";
      }
      return true;
    }
  }

  class ConstraintSeq {
    var $conn;
      
    var $exps; //collection of expression objects
    var $ops;  //collection of operators between expressions: $exps[i] $ops[i] $exps[i+1] ...
    
    function ConstraintSeq($conn) {
      $this->conn = $conn;
    }
    
    /* 
        $cseq - input
        $rlns - set of used relation (for validation check)
    */
    function parse($cseq, &$rlns, $otid) {
      $cexs = Array();
      $ops = Array();
      
      $lp = 0;
      $lp_pos = 0;
      $rp = 0;
      $len = StrLen($cseq);
      $pos = 0;            
      $last_seg_pos = 0;
      if($cseq[0] != "(") {
        syn_err(sprintf("HAVING elements must be enclosed into parentheses [%s]", $cseq));
        return false;
      }
      while($pos < $len) {
        if($cseq[$pos] == "(") {
          if($lp == 0)      //segment begin
            $lp_pos = $pos;
          $lp++;
        }
        else if($cseq[$pos] == ")") {
          $rp++;
        }
        if($lp == $rp && $lp > 0) { //segment found
          $lp = $rp = 0; //reset
          $last_seg_pos = $pos;
          $cexs[] = Trim(SubStr($cseq, $lp_pos + 1, $pos - $lp_pos - 1));
          /* try to get operator */
          $rest = Trim(SubStr($cseq, $pos + 1));
          if(StrLen($rest) > 0) {
            $op = Trim(StrToUpper(StrTok($rest, " (")));
            if($op != "AND" && $op != "OR") {
              syn_err(sprintf("Invalid operator [%s] in [%s]", $op, $cseq));
              return false;
            }
            $ops[] = $op;
          }
        }
        $pos++;
      }
      if($lp != $rp) {
        syn_err(sprintf("Invalid number of matching parenthesis in [%s]", $cseq));
        return false;
      }
      if($last_seg_pos < $len-1) {
        syn_err(sprintf("No valid constraint found in [%s]", SubStr($cseq, $last_seg_pos + 1)));
        return false;
      }
      /* create array of constraint expressions */
      $this->exps = Array();
      for($i=0; $i<Count($cexs); $i++) {
        $exp = new ConstraintExpr($this->conn);
        if(!$exp->parse($cexs[$i], $rlns, $otid))
          return false;
        $this->exps[] = $exp;
      }
      /* assign array of constraint operators */
      $this->ops = $ops;
      
//      debug_trace("[".$cseq."] printing");
//      print_array($cexs);
//      print_array($ops);
      
      return true;
    }
    
    function dumpParsed() {
      $res = "";
      for($i=0; $i<Count($this->exps); $i++) {
        $el = $this->exps[$i];
        $res .= $el->dumpParsed();
        if($i < Count($this->ops)) {
          $op = $this->ops[$i];
          $res .= " ".$op." ";
        }
      }
      return $res;
    }
    
    function buildSelect($uplink, &$cnt, &$from, &$where) {
      for($i=0; $i<Count($this->exps); $i++) {
        $el = $this->exps[$i];
        if(!$el->buildSelect($uplink, $cnt, $from, $where))
          return false;
        if($i < Count($this->ops)) {
          $op = $this->ops[$i];
          $where .= " ".$op." ";
        }
      }
      return true;      
    }
  }
?>