<?
  
  /*
    contains:
    - Join class
    - JoinExpr class
    - JoinSeq  class
  */

  class Join {
    var $conn;
      
    var $not;      //"NOT" or ""
    var $otid;     //object type_id parent is connected with
    var $rev;      //true if relationship $rel_id navigates in reverse order
    var $relid;   //relationship ID
    var $hv;       //HAVING part (optional)
    var $jd;       //JOIN part (optional)

    
    function Join($conn) {
      $this->conn = $conn;
    }

    /* 
        $initial - input
        $rlns - set of used relations (for validation check)
        $rels - set of used relationships (for validation check)
        $masterOtid - master side type_id
    */
    function parseInitial($initial, &$rlns, &$rels, $masterOtid) {
      $words = split("[ ]+", Trim($initial));
      
      $not;
      $with;
      $otname;
      $through;
      $rev;
      $relname;
      
      /* 4 - 6 words */
      switch(Count($words)) {
        case 4: //WITH, otname, THROUGH, relname
          $not = "";
          $with = StrToUpper($words[0]);
          $otname = $words[1];
          $through = StrToUpper($words[2]);
          $rev = false;
          $relname = $words[3];
          break;
        case 5: //NOT, WITH, otname, THROUGH, relname | WITH, otname, THROUGH, REV, relname
          if(StrToUpper($words[0]) == "NOT") {
            $not = "NOT";
            $with = StrToUpper($words[1]);
            $otname = $words[2];
            $through = StrToUpper($words[3]);
            $rev = false;
            $relname = $words[4];
          }
          else if(StrToUpper($words[3]) == "REV") {
            $not = "";
            $with = StrToUpper($words[0]);
            $otname = $words[1];
            $through = StrToUpper($words[2]);
            $rev = true;
            $relname = $words[4];
          }
          else {
            syn_error(sprintf("Invalid join syntax [%s]", $initial));
            return false;
          }
          break;
        case 6: //NOT, WITH, otname, THROUGH, REV, relname
          $not = StrToUpper($words[0]);
          $with = StrToUpper($words[1]);
          $otname = $words[2];
          $through = StrToUpper($words[3]);
          $rev = StrToUpper($words[4]);
          $relname = $words[5];
          if($not != "NOT") {
            syn_err(sprintf("found [%s], expected [%s] in [%s]", $not, "NOT", $initial));
            return false;
          }
          if($rev != "REV") {
            syn_err(sprintf("found [%s], expected [%s] in [%s]", $rev, "REV", $initial));
            return false;
          }
          else
            $rev = true;
          break;
        default:
//          print_array($words);
          syn_err(sprintf("Invalid join syntax [%s]", $initial));
          return false;
      }
      
      /* global controls & pre-evaluation */
      if($with != "WITH") {
        syn_err(sprintf("found [%s], expected [%s] in [%s]", $with, "WITH", $initial));
        return false;
      }
      if($through != "THROUGH") {
        syn_err(sprintf("found [%s], expected [%s] in [%s]", $with, "THROUGH", $initial));
        return false;
      }

      //get type_id
      $otid = getTypeId($this->conn, $otname);
      if(!$otid)
        return false;

      //get relationship id
      $relid =  getRspId($this->conn, $relname, $rev);
      if(!$relid)
        return false;

      //check if given relationship is legal for this object type
      if($rev) {
        if(!isLegal($this->conn, $relid, $otid, $masterOtid, $rev))
          return false;
      }
      else {
        if(!isLegal($this->conn, $relid, $masterOtid, $otid, $rev))
          return false;
      }
        
      $this->not = $not;
      $this->otid = $otid;
      $this->rev = $rev;
      $this->relid = $relid;
      
      $rels[$this->relid] = 1;
      
      return true;      
    }
    
    /* 
        $constr - input
        $rlns - set of used relations (for validation check)
        $rels - set of used relationships (for validation check)
    */
    function parse($constr, &$rlns, &$rels, $masterOtid) {
      /* separate parts (INITIAL, HAVING, JOINED)*/
      
      $hKey = " HAVING ";
      $jKey = " JOIN ";
      
//      debug_trace($constr);
      
      $hpos = StrPos(StrToUpper($constr), $hKey);
      $jpos = StrPos(StrToUpper($constr), $jKey);
      
      if($jpos > 0 && $hpos > $jpos) {
        //probably has "HAVING" as part of "JOIN" change:
        $hpos = 0;
      }
      
      $initial = "";
      $having = "";
      $joined = "";
      
      if(!$hpos && !$jpos)  //neither HAVING nor JOIN are specified
        $initial = Trim($constr);
      else if(!$hpos){ //only JOIN
        $initial = Trim(Substr($constr, 0, $jpos));
        $joined = Trim(SubStr($constr, $jpos + StrLen($jKey)));
      }
      else { //both or only HAVING
        $initial = Trim(Substr($constr, 0, $hpos));
        if(!$jpos) //only HAVING
          $having = Trim(SubStr($constr, $hpos + StrLen($hKey)));
        else { //both
          $having = Trim(SubStr($constr, $hpos + StrLen($hKey), $jpos - $hpos - StrLen($hKey)));
          $joined = Trim(SubStr($constr, $jpos + StrLen($jKey)));    
        }
      }
      if(StrLen($initial) == 0) {
        syn_err(sprintf("&lt;WITH ... THROUGH&gt; part of the join statement [%s] is empty", $constr));
        return false;
      }
      if($hpos && StrLen($having) == 0) {
        syn_err(sprintf("&lt;having&gt; part the join statement [%s] is empty", $constr));
        return false;
      }
      if($jpos && StrLen($joined) == 0) {
        syn_err(sprintf("&lt;joined&gt; part the join statement [%s] is empty", $constr));
        return false;
      }

//      debug_trace("initial: ".$initial);
//      debug_trace("having: ".$having);
//      debug_trace("joined: ".$joined);

      /* WITH_THROUGH (not recursive) */
      if(!$this->parseInitial($initial, $rlns, $rels, $masterOtid))
        return false;

      /* HAVING PART */
      if($having == "") {
        $this->hv = false;
      }
      else {
        $this->hv = new ConstraintSeq($this->conn);
        if(!$this->hv->parse($having, $rlns, $this->otid))
          return false;  
      }

      /* JOINED PART */          
      if($joined == "") {
        $this->jd = false;
      }
      else {
        $this->jd = new JoinSeq($this->conn);
        if(!$this->jd->parse($joined, $rlns, $rels, $this->otid))
          return false;
      }
        
      return true;
    }
    
    function dumpParsed() {
       $result = "";
       /* dump WITH_THROUGH */
       $result .= sprintf("%sWITH type_id(%d) THROUGH %srel_id(%d)",
         $this->not ? "NOT " : "",
         $this->otid,
         $this->rev ? "REV " : "",
         $this->relid);
       
       /* dump HAVING */
       if($this->hv) {
         $result .= " HAVING ";
         $result .= $this->hv->dumpParsed();
       }
       
       /* dump JOINED */
       if($this->jd) {
         $result .= " JOIN ";
         $result .= $this->jd->dumpParsed();
       }
       
       return $result;
    }    
    
    function buildSelect($uplink, &$cnt, &$from, &$where) {
      $relta = "RELC".($cnt++);
      $ota = "OBJ".($cnt++);

      $where .= sprintf("%s.OID %s IN ",$uplink, $this->not);
      if($this->rev) {
        $where .= sprintf("(SELECT %s.OID_R FROM RELATIONSHIPS_CACHE %s, OBJECTS %s",
          $relta, $relta, $ota);
        $where .= sprintf(" WHERE %s.REL_ID=%d AND %s.TYPE_ID=%d AND %s.OID_L=%s.OID",
          $relta, $this->relid, $ota, $this->otid, $relta, $ota);
      }
      else {
        $where .= sprintf("(SELECT %s.OID_L FROM RELATIONSHIPS_CACHE %s, OBJECTS %s",
          $relta, $relta, $ota);
        $where .= sprintf(" WHERE %s.REL_ID=%d AND %s.TYPE_ID=%d AND %s.OID_R=%s.OID",
          $relta, $this->relid, $ota, $this->otid, $relta, $ota);
      }
      
      /* build having */
      if($this->hv) {
        $where .= " AND (";
        if(!$this->hv->buildSelect($ota, $cnt, $from, $where))
          return false;
        $where .= ")";
      }
      
      /* build joined */
      if($this->jd) {
        $where .= " AND (";
        if(!$this->jd->buildSelect($ota, $cnt, $from, $where))
          return false;
        $where .= ")";
      }
      
      $where .= ")";
      return true;
    }
  }

  class JoinExpr {
    var $conn;
      
    var $cont;  //a single constraint
    var $cseq;  //a constraint sequence

    function JoinExpr($conn) {
      $this->conn = $conn;
    }

    /* 
        $cexp - input
        $rlns - set of used relations (for validation check)
        $rels - set of used relationships (for validation check)
    */
    function parse($cexp, &$rlns, &$rels, $masterOtid) {
      if(eregi("[ \)]AND[ \(]", $cexp) || eregi("[ \)]OR[ \(]", $cexp)) {
        $cseq = new JoinSeq($this->conn);
        if(!$cseq->parse($cexp, $rlns, $rels, $masterOtid))
          return false;
        $this->cseq = $cseq;
        $this->cont = false;
      }
      else {
        $cont = new Join($this->conn);
        if(!$cont->parse($cexp, $rlns, $rels, $masterOtid))
          return false;
        $this->cseq = false;
        $this->cont = $cont;
      }
      return true;
    }
    
    function dumpParsed() {
      if($this->cseq)
        return "(".$this->cseq->dumpParsed().")";
      else
        return "(".$this->cont->dumpParsed().")";
    }

    function buildSelect($uplink, &$cnt, &$from, &$where) {
      if($this->cseq) {
        $where .= "(";
        $this->cseq->buildSelect($uplink, $cnt, $from, $where);
        $where .= ")";
      }
      else {
        $where .= "(";
        $this->cont->buildSelect($uplink, $cnt, $from, $where);
        $where .= ")";
      }
      return true;
    }
  }

  class JoinSeq {
    var $conn;
      
    var $exps; //collection of join expression objects
    var $ops;  //collection of operators between expressions: $exps[i] $ops[i] $exps[i+1] ...
    
    function JoinSeq($conn) {
      $this->conn = $conn;
    }
    
    /* 
        $cseq - input
        $rlns - set of used relations (for validation check)
        $rels - set of used relationships (for validation check)
    */
    function parse($cseq, &$rlns, &$rels, $masterOtid) {
      $cexs = Array();
      $ops = Array();
      
      $lp = 0;
      $lp_pos = 0;
      $rp = 0;
      $len = StrLen($cseq);
      $pos = 0;            
      $last_seg_pos = 0;
      if($cseq[0] != "(") {
        syn_err(sprintf("JOIN elements must be enclosed into parentheses [%s]", $cseq));
        return false;
      }
      while($pos < $len) {
        if($cseq[$pos] == "(") {
          if($lp == 0)      //segment begin
            $lp_pos = $pos;
          $lp++;
        }
        else if($cseq[$pos] == ")") {
          $rp++;
        }
        if($lp == $rp && $lp > 0) { //segment found
          $lp = $rp = 0; //reset
          $last_seg_pos = $pos;
          $cexs[] = Trim(SubStr($cseq, $lp_pos + 1, $pos - $lp_pos - 1));
          /* try to get operator */
          $rest = Trim(SubStr($cseq, $pos + 1));
          if(StrLen($rest) > 0) {
            $op = Trim(StrToUpper(StrTok($rest, " (")));
            if($op != "AND" && $op != "OR") {
              syn_err(sprintf("Invalid operator [%s] in [%s]", $op, $cseq));
              return false;
            }
            $ops[] = $op;
          }
        }
        $pos++;
      }
      if($lp != $rp) {
        syn_err(sprintf("Invalid number of matching parenthesis in [%s]", $cseq));
        return false;
      }
      if($last_seg_pos < $len-1) {
        syn_err(sprintf("No valid join found in [%s]", SubStr($cseq, $last_seg_pos + 1)));
        return false;
      }
      /* create array of constraint expressions */
      $this->exps = Array();
      for($i=0; $i<Count($cexs); $i++) {
        $exp = new JoinExpr($this->conn);
        if(!$exp->parse($cexs[$i], $rlns, $rels, $masterOtid))
          return false;
        $this->exps[] = $exp;
      }
      /* assign array of constraint operators */
      $this->ops = $ops;
      
//      debug_trace("[".$cseq."] printing");
//      print_array($cexs);
//      print_array($ops);
      
      return true;
    }
    
    function dumpParsed() {
      $res = "";
      for($i=0; $i<Count($this->exps); $i++) {
        $el = $this->exps[$i];
        $res .= $el->dumpParsed();
        if($i < Count($this->ops)) {
          $op = $this->ops[$i];
          $res .= " ".$op." ";
        }
      }
      return $res;
    }
    
    function buildSelect($uplink, &$cnt, &$from, &$where) {
      for($i=0; $i<Count($this->exps); $i++) {
        $el = $this->exps[$i];
        if(!$el->buildSelect($uplink, $cnt, $from, $where))
          return false;
        if($i < Count($this->ops)) {
          $op = $this->ops[$i];
          $where .= " ".$op." ";
        }
      }
      return true;      
    }
  }
?>