<?

  /* main QL file
     contains:
     - SQLTemplate class
     - MQLEngine class
     - MCUBE_Exec function
     note:
     SELECT PART of query is very simple, it is parsed
     and built in MQLEngine class.
  */
  
  require "./helper.php";    //helper functions
  require "./having_ql.php"; //HAVING PART objects (parsing, SQL generating)
  require "./joined_ql.php"; //JOINED PART objects (parsing, SQL generating)
  require "../rel_types/rsp_cache.php"; //relationship closure build
  require "../relation_types/relation_cache.php"; //relation closure build
  
  class SQLTemplate {
    var $select;
    var $from;
    var $where;
    var $order_by;
    
    function SQLTemplate() {
      $this->select = "SELECT DISTINCT OBJ0.OID, OBJ0.TYPE_ID, OBJ0.ONAME";
      $this->from = " FROM OBJECTS OBJ0";
      $this->where = " WHERE OBJ0.TYPE_ID=";
      $this->order_by = " ORDER BY OBJ0.ONAME";
    }
    
    function toString() {
       $command = $this->select;
       $command .= $this->from;
       $command .= $this->where;
       $command .= $this->order_by;
      return $command;
    }
  }

  class MQLEngine {
    var $conn;
    
    /* internals */
    var $otid;       //(SELECT PART) type_id of object type to search for
    var $hv;         //(HAVING PART) (top level)
    var $jd;         //(JOINED PART) (top level)
    
    var $sqlt;       //SQL template object
    var $rlns;       //relations used in query
    var $rels;       //relationships used in query
    
    function MQLEngine($conn) {
      $this->conn = $conn;
    }
   
    function parseSelect($select) {
      $otid = getTypeId($this->conn, $select);
      if(!$otid)
        return false;
      else {
        $this->otid = $otid;
        return true;
      }
    }

    function parseQuery($mqlCommand) {
      /* separate parts (SELECT, HAVING, JOINED)*/
      $mqlCommand = StrTr($mqlCommand, "\n\t", "  ");
      $mqlCommand = Trim($mqlCommand);
      
      $sKey = "SELECT";
      $hKey = " HAVING ";
      $jKey = " JOIN ";
      
//      debug_trace($mqlCommand);
      
      if(StrToUpper(SubStr($mqlCommand, 0, StrLen($sKey))) != $sKey) {
        syn_err("Missing \"SELECT\" keyword");
        return false;
      }
      
      $hpos = StrPos(StrToUpper($mqlCommand), $hKey);
      $jpos = StrPos(StrToUpper($mqlCommand), $jKey);
      
      if($jpos > 0 && $hpos > $jpos) {
        //probably has "HAVING" as part of "JOIN" change:
        $hpos = 0;
//        syn_err("HAVING must precede JOIN");
//        return false;
      }
      
      $select = "";
      $having = "";
      $joined = "";
      
      if(!$hpos && !$jpos)  //neither HAVING nor JOIN are specified
        $select = Trim(SubStr($mqlCommand, StrLen($sKey)));
      else if(!$hpos){ //only JOIN
        $select = Trim(Substr($mqlCommand, StrLen($sKey), $jpos - StrLen($sKey)));
        $joined = Trim(SubStr($mqlCommand, $jpos + StrLen($jKey)));
      }
      else { //both or only HAVING
        $select = Trim(Substr($mqlCommand, StrLen($sKey), $hpos - StrLen($sKey)));
        if(!$jpos) //only HAVING
          $having = Trim(SubStr($mqlCommand, $hpos + StrLen($hKey)));
        else { //both
          $having = Trim(SubStr($mqlCommand, $hpos + StrLen($hKey), $jpos - $hpos - StrLen($hKey)));
          $joined = Trim(SubStr($mqlCommand, $jpos + StrLen($jKey)));    
        }
      }
      if(StrLen($select) == 0) {
        syn_err("&lt;typename&gt; part of SELECT is empty");
        return false;
      }
      if($hpos && StrLen($having) == 0) {
        syn_err("&lt;having&gt; part of SELECT is empty");
        return false;
      }
      if($jpos && StrLen($joined) == 0) {
        syn_err("&lt;joined&gt; part of SELECT is empty");
        return false;
      }
/*
      debug_trace("select: ".$select);
      debug_trace("having: ".$having);
      debug_trace("joined: ".$joined);
*/
      /* SELECT PART (not recursive) */
      if(!$this->parseSelect($select))
        return false;

      /* init state variables */
      $this->rlns = Array();
      $this->rels = Array();
              
      /* HAVING PART */
      if($having == "") {
        $this->hv = false;
      }
      else {
        $this->hv = new ConstraintSeq($this->conn);
        if(!$this->hv->parse($having, $this->rlns, $this->otid))
          return false;  
      }

      /* JOINED PART */          
      if($joined == "") {
        $this->jd = false;
      }
      else {
        $this->jd = new JoinSeq($this->conn);
        return $this->jd->parse($joined, $this->rlns, $this->rels, $this->otid);  
      }
        
      return true;
    }
     
     function rebuildClosures($showRebuild) {
       if($showRebuild) {
         echo "<P><B>Rebuilding relations:</B><BR>\n";
       }
       $cnt = Count($this->rlns);
       while($cnt--) {
         if($showRebuild) {echo Key($this->rlns)." ... ";}
         $rc = new RelationCache($this->conn, Key($this->rlns));
         $rc->validate($showRebuild);
         Next($this->rlns);
       }
       if($showRebuild) {
         echo "<P><B>Rebuilding relationships:</B><BR>\n";
       }
       $cnt = Count($this->rels);
       while($cnt--) {
         if($showRebuild) {echo Key($this->rels)." ... ";}
         $rc = new RspCache($this->conn, Key($this->rels));
         $rc->validate($showRebuild);
         Next($this->rels);
       }
       if($showRebuild) {
         echo "<BR>";
       }
       return true;
     }
     
     function dumpParsed() {
       $result = "";
       /* dump SELECT */
       $result .= sprintf("SELECT type_id(%d)\n", $this->otid);
       
       /* dump HAVING */
       if($this->hv) {
         $result .= "<BR>HAVING ";
         $result .= $this->hv->dumpParsed();
       }
       
       /* dump JOINED */
       if($this->jd) {
         $result .= "<BR>JOIN ";
         $result .= $this->jd->dumpParsed();
       }
       
       return $result;
     }
     
     function buildSelect() {
       $this->sqlt->where .= $this->otid;
       return true;
     }
       
     function buildQuery() {
       $this->sqlt = new SQLTemplate();
     
       if(!$this->buildSelect())
         return false;

      $uplink = "OBJ0";
      $cnt = 1;
      $from = "";
      $where = "";
      /* build having */
      if($this->hv) {
        if(!$this->hv->buildSelect($uplink, $cnt, $from, $where))
          return false;
        $this->sqlt->from .= $from;
        $from = "";
        $this->sqlt->where .= " AND (".$where.")";
        $where = "";
      }
      
      /* build joined */
      if($this->jd) {
        if(!$this->jd->buildSelect($uplink, $cnt, $from, $where))
          return false;
        $this->sqlt->from .= $from;
        $from = "";
        $this->sqlt->where .= " AND (".$where.")";
        $where = "";
      }

      return $this->sqlt->toString();
     }

     function translateToSQL($mqlCommand, $flag) {
       if($this->parseQuery($mqlCommand)) {
       
         if($flag == "exec") {
           global $showRebuild;
           if(!$this->rebuildClosures($showRebuild)) {
             return false;
           }
         }
         if($flag == "parse")
           return $this->dumpParsed();
         else
           return $this->buildQuery();
       }
       else {
         return false;
       }
     } 
  }


  function MCUBE_Exec($conn, $mqlCommand, $flag="exec") {
    $mqle = new MQLEngine($conn);
    $sqlCommand = $mqle->translateToSQL($mqlCommand, $flag);
    if(!$sqlCommand)
      return;  //errors are reported by engine
      
    if($flag == "plan" || $flag == "parse") {
      return $sqlCommand;
    }
    else {
      return ODBC_Exec($conn, $sqlCommand);
    }
  }

?>